const {yellow, green, red} = require('chalk');
const path = require('path');
const createRollupConfig = require('../config/rollup.config.default');
const rollup = require('rollup');
const express = require('express');
const mashupRollupPlugin = require('../plugins/mashupRollupPlugin');
const cors = require('cors');

const runRollupDevServer = async function(
  dir = process.cwd(),
  {minify = false, dependencyReplace, outputDir, port = 9090}
) {
  outputDir = outputDir || path.join(dir, 'dist');
  const basePath = `http://localhost:${port}`;
  const config = await createRollupConfig({
    dir,
    minify,
    dependencyReplace,
    outputDir,
    basePath
  });
  const {
    settings: {mashupModuleName}
  } = config.plugins.find(
    plugin => plugin.name === mashupRollupPlugin.pluginName
  );
  const server = express();
  server.use(cors());

  let complete = false;
  let completionListeners = [];
  const waitUntilBuildComplete = (fn) => {
    if (complete) {
      fn();
    } else {
      completionListeners.push(fn);
    }
  };
  const setBuildComplete = (value) => {
    complete = value;
    if (complete) {
      completionListeners.forEach((fn) => fn());
      completionListeners = [];
    }
  };
  let buildErrorMessage = null;

  const serveStatic = express.static(outputDir);
  server.use(function(req, res, next) {
    waitUntilBuildComplete(function() {
      if (buildErrorMessage && req.url.startsWith('/index.js')) {
        res.send(`console.error(${JSON.stringify(mashupModuleName)} + " build failed:\\n" + ${JSON.stringify(buildErrorMessage)});`);
      } else {
        serveStatic(req, res, next);
      }
    });
  });
  server.set('view engine', 'pug');
  server.set('views', path.join(__dirname, '../config'));
  server.use((request, response, next) => {
    if (request.url === '/') {
      const bundleAbsoluteLink = config.input[0]
      const bundleExtension = path.extname(bundleAbsoluteLink)

      response.render('index', {
        name: 'create-mashup-app dev server',
        moduleName: mashupModuleName,
        bundleLink: path.basename(bundleAbsoluteLink, bundleExtension) + '.js'
      });
    } else {
      next();
    }
  });
  let serverStarting = false;
  let firstBuildPassed = false;

  const watcher = rollup.watch(config);
  watcher.on('event', event => {
    switch (event.code) {
      case 'START': {
        if (!serverStarting) {
          serverStarting = true;
          server.listen(port);
        }
        break;
      }
      case 'BUNDLE_START': {
        console.log(
          `Building ${yellow(event.input[0])} -> ${green(event.output[0])}...`
        );
        setBuildComplete(false);
        buildErrorMessage = null;
        break;
      }
      case 'END': {
        if (!firstBuildPassed) {
          firstBuildPassed = true;
          console.log(`Server listening at ${yellow(basePath)}`);
        }
        setBuildComplete(true);
        break;
      }
      case 'ERROR': {
        console.log(red(event.error.message));
        console.log(red(event.error.stack));
        buildErrorMessage = event.error.message;
        setBuildComplete(true);
        break;
      }
      case 'FATAL': {
        console.log(red(event.error.message));
        // TODO upgrade rollup because modern version has no FATAL event (thus seems to be more stable)
        // FATAL event happens now (why?) e.g. when first build is erroneous, nothing is being watched after that, manual restart is required
        process.exit(1);
      }
    }
  });
};

module.exports = async (...args) => {
  try {
    await runRollupDevServer(...args);
  } catch (err) {
    console.error(red(err));
    console.error(err.stack);
    process.exit(-1);
  }
};
