const path = require('path');
const {red, green, cyan, yellow} = require('chalk');
const {identity} = require('lodash');
const cosmiconfig = require('cosmiconfig');
const errors = require('./errors.json');
const {babel} = require('@rollup/plugin-babel');
const postcss = require('rollup-plugin-postcss');
const json = require('@rollup/plugin-json');
const commonjs = require('@rollup/plugin-commonjs');
const {nodeResolve: resolve} = require('@rollup/plugin-node-resolve');
const iife = require('rollup-plugin-iife');
const {terser} = require('rollup-plugin-terser');
const typescript2 = require('rollup-plugin-typescript2');
const replace = require('@rollup/plugin-replace');
const mashup = require('../plugins/mashupRollupPlugin');
const chunkInfo = require('../plugins/chunkInfoPlugin');
const cleanOutput = require('../plugins/cleanPlugin');
const includePaths = require('rollup-plugin-includepaths');

async function getPostcssConfig(dir, getVariables = () => ({})) {
  const postcssConfigSearchResult = await cosmiconfig('postcss').search(dir);
  const postcssConfig = (postcssConfigSearchResult &&
    postcssConfigSearchResult.config) || {
    plugins: [
      require('postcss-import'),
      require('postcss-global-import'),
      require('postcss-each'),
      require('postcss-mixins'),
      require('postcss-simple-vars')({
        variables: () => {
          const vars = getVariables();
          return vars || {};
        }
      }),
      require('postcss-nested-ancestors'),
      require('postcss-nested'),
      require('postcss-current-selector'),
      require('postcss-color-function'),
      require('postcss-calc'),
      require('postcss-inline-svg'),
      require('autoprefixer')
    ]
  };
  return postcssConfig;
}

module.exports = async ({
  dir = process.cwd(),
  minify,
  dependencyReplace = true,
  target,
  outputDir,
  basePath
}) => {
  const configSearchResult = await cosmiconfig('mashup').search(dir);
  if (!configSearchResult) {
    throw new Error(errors.configNotFound);
  }

  const {config: mashupConfig, filepath} = configSearchResult;

  const entryPoint = path.resolve(dir, mashupConfig.entry || 'src/index.js');
  const publicApiEntry = mashupConfig.publicApi
    ? path.resolve(dir, mashupConfig.publicApi)
    : null;
  const targetPackageJson = require(path.join(dir, 'package.json'));
  let configTransform = mashupConfig.configTransform;
  basePath = mashupConfig.basePath || basePath;

  console.log(` - loaded mashup config from ${cyan(filepath)}`);
  console.log(
    ` - minification is ${minify ? green('enabled') : red('disabled')}`
  );
  console.log(` - entry point: ${cyan(entryPoint)}`);

  if (publicApiEntry) {
    console.log(` - api entry point: ${cyan(publicApiEntry)}`);
  }

  if (typeof configTransform === 'function') {
    console.log(
      ' - using custom config transform (function declared in config)'
    );
  } else if (typeof configTransform === 'string') {
    const configTransformFilePath = path.join(dir, configTransform);
    console.log(
      ` - using custom config transform from ${cyan(configTransformFilePath)}`
    );
    configTransform = require(configTransformFilePath);
  }
  if (typeof configTransform !== 'function') {
    configTransform = identity;
  }

  const mashupName = mashupConfig.name || targetPackageJson.name;
  const mashupVersion = targetPackageJson.version;
  const mashupModuleName = mashupConfig.moduleName || mashupName;
  const mashupIntegrationName = mashupConfig.integrationName || mashupName;
  const mashupEntryFileName  = mashupConfig.mashupEntryFileName || 'index.js';
  const {dependencies = []} = mashupConfig;

  // collect all dependencies that should be excluded from the bundle and
  // instead taken from global object
  const externalDependencies = dependencies.map(
    dependency =>
      dependency.provideModule || dependency.moduleName || dependency
  );
  // all external dependencies are dumped into one object on window, and can be taken from there
  // we also need unique name for this particular external, so we generate it with
  // a bit of randomness on each build
  const randomString = Math.random()
    .toString(36)
    .substring(2, 7);
  const mashupGlobalName = `__${mashupName.replace(
    /[- ]/g,
    ''
  )}${mashupVersion.replace(/[\.\+-]/g, '')}_${randomString}`;
  const mashupDependenciesGlobalName = `${mashupGlobalName}_deps`;
  const mashupChunksGlobalName = `${mashupGlobalName}_chunks`;
  const mashupImportGlobalName = `${mashupGlobalName}_import`;
  const mashupDependencyMapGlobalName = `${mashupGlobalName}_depMap`;
  const mashupResolveUrlGlobalName = `${mashupGlobalName}_resolveUrl`;
  const mashupInitializeFunction = `${mashupGlobalName}_initialize`;
  if (basePath && !mashupConfig.integrationName) {
    console.warn(
      yellow(
        `Warning: detected basePath, but no explicit integrationName in mashup config; ` +
          `"${mashupIntegrationName}" will be used as integration name instead. ` +
          `basePath means you are trying to do code splitting; to correctly version ` +
          `your mashup chunks, provide integrationName according to your docker image config so ` +
          `that your chunk URLs are correctly versioned by tau.mashups.resolveIntegrationUrl. ` +
          `See https://gitlab.tpondemand.net/frontend/create-mashup-app/#code-splitting for details. `
      )
    );
  }
  outputDir = outputDir ? path.resolve(outputDir) : path.join(dir, 'dist');

  const cssVariablesConfig = mashupConfig.cssVariables;
  const getCssVariables = () => {
    if (typeof cssVariablesConfig === 'string') {
      return require(cssVariablesConfig);
    } else if (typeof cssVariablesConfig === 'function') {
      return cssVariablesConfig();
    }

    return cssVariablesConfig;
  };

  const getChunkName = chunkInfo => chunkInfo.modules[entryPoint] ? mashupEntryFileName : '[name]-[hash].js';

  const rollupConfig = {
    input: [entryPoint, ...(publicApiEntry ? [publicApiEntry] : [])],
    output: {
      dir: outputDir,
      format: 'es',
      entryFileNames: getChunkName,
      chunkFileNames: getChunkName,
      name: mashupGlobalName,
      sourcemap: !minify,
      globals: localDependencyName => {
        if (localDependencyName.startsWith(outputDir)) {
          const chunkName = localDependencyName
            .replace(outputDir, '')
            .substring(1);
          // all code-split chunks are written as keys of <something>_chunks object
          return `window.${mashupChunksGlobalName}['./${chunkName}']`;
        }
        // all other modules are externals (i.e. react) and are extracted from
        // keys of <something>_deps
        return `window.${mashupDependenciesGlobalName}['${localDependencyName}']`;
      }
    },
    external: dependencyReplace ? externalDependencies : null,
    plugins: [
      resolve({
        mainFields: ['browser', 'module', 'main'],
        extensions: ['.js', '.jsx']
      }),
      // typescript, babel and commonjs can usually perform DCE based
      // on defined replaced strings, so replace goes before them
      replace({
        'process.env.NODE_ENV': JSON.stringify(
          minify ? 'production' : 'development'
        )
      }),
      includePaths({
        paths: [dir],
        extensions: ['.js', '.jsx', '.ts', '.tsx']
      }),
      typescript2({
        jsx: 'React',
        include: /\.tsx?/,
        cacheRoot: './node_modules/.rpt2_cache',
        tsconfigDefaults: {
          compilerOptions: {
            allowSyntheticDefaultImports: true,
            module: 'es2015',
            target: 'es6',
            jsx: 'react'
          }
        }
      }),
      json(),
      babel({
        babelHelpers: 'runtime',
        exclude: /node_modules(\/|\\)(.*)/,
        extensions: ['.js', '.jsx', '.ts', '.tsx', '.json'],
        presets: [
          [
            '@babel/preset-env',
            {
              modules: false,
              useBuiltIns: 'usage',
              corejs: 3,
              targets: {
                browsers: targetPackageJson.browserlist || [
                  'last 2 Chrome versions',
                  'last 2 Firefox versions',
                  'last 2 Safari versions'
                ]
              }
            }
          ],
          '@babel/preset-react'
        ],
        plugins: [
          ['@babel/plugin-transform-runtime', {corejs: 3}],
          '@babel/plugin-syntax-dynamic-import',
          '@babel/plugin-proposal-object-rest-spread',
          '@babel/plugin-proposal-class-properties',
          '@babel/plugin-transform-property-literals',
          '@babel/plugin-transform-member-expression-literals'
        ]
      }),
      commonjs({
        namedExports: targetPackageJson.namedExports,
        transformMixedEsModules: true
      }),
      postcss({
        modules: {
          generateScopedName: className => {
            return `${className}__${mashupName}__${mashupVersion
              .split('.')
              .join('-')}`
          },
          localsConvention: 'camelCase'
        },
        inject: true,
        autoModules: false,
        // babel prioritizes config file (.babelrc) over hardcoded config
        // but postcss doesn't, so we need to manually search for custom config file
        // luckily, it uses cosmiconfig as well
        ...(await getPostcssConfig(dir, getCssVariables))
      }),
      iife(),
      mashup({
        basePath,
        entryPoint,
        dependencies,
        target,
        mashupModuleName,
        mashupGlobalName,
        mashupChunksGlobalName,
        mashupImportGlobalName,
        mashupDependenciesGlobalName,
        mashupDependencyMapGlobalName,
        mashupResolveUrlGlobalName,
        mashupIntegrationName,
        mashupInitializeFunction,
        publicApiEntry
      }),
      chunkInfo(),
      cleanOutput({inputDir: dir})
    ]
  };

  if (minify) {
    rollupConfig.plugins.push(terser());
  }

  return configTransform(rollupConfig);
};
