const {walk} = require('estree-walker');
const {generate} = require('astring');
const MagicString = require('magic-string');
const Concat = require('concat-with-sourcemaps');
const {cyan} = require('chalk');

const PLUGIN_NAME = 'mashup-entry-plugin';

/**
 * Installs a dynamic module loader for this mashup.
 * It will create a new global with name `mashupImportGlobalName`
 * and initialize it with a function that loads another chunk
 * and returns a promise for that chunk. It will resolve immediately
 * if the chunk is already loaded on the page.
 */
function getMashupModuleLoader({
  mashupGlobalName,
  mashupDependenciesGlobalName,
  mashupChunksGlobalName,
  mashupImportGlobalName,
  mashupDependencyMapGlobalName,
  mashupResolveUrlGlobalName,
  mashupIntegrationName,
  dependencyMap,
  basePath,
  dependencies
}) {
  return `
window.${mashupChunksGlobalName} = window.${mashupChunksGlobalName} || {};
window.${mashupDependencyMapGlobalName} = ${JSON.stringify(dependencyMap)};
window.${mashupDependenciesGlobalName} = {};
window.${mashupGlobalName}_getDependencies = function () {
  return tau.mashups${dependencies
    .map(item => `.addDependency('${item.mashupDependency || item}')`)
    .join('\n')};
};
window.${mashupGlobalName}_evaluated = false;

window.${mashupResolveUrlGlobalName} = (function(resolveIntegrationUrl) {
    return function(url) {
        if (url.startsWith('./')) {
            url = url.substring(2);
        }
        var fallbackUrl = '${basePath}' + '/' + url;
        return typeof resolveIntegrationUrl === 'function'
            ? resolveIntegrationUrl('${mashupIntegrationName}', fallbackUrl)
            : fallbackUrl;
    }
})(tau.mashups && tau.mashups.resolveIntegrationUrl)
window.${mashupImportGlobalName} = (function(moduleId) {
    if (!moduleId.startsWith('./')) {
        moduleId = './' + moduleId;
    }
    
    const pending = this;
    
    if (window.${mashupChunksGlobalName}[moduleId]) {
      return Promise.resolve(window.${mashupChunksGlobalName}[moduleId]);
    }
        
    if (!pending[moduleId]) {
        pending[moduleId] = {};
        
        const dependencies = window.${mashupDependencyMapGlobalName} && window.${mashupDependencyMapGlobalName}[moduleId] || []
        const dependencyPromises = dependencies.map(function(dependencyModuleId) {
            return window.${mashupImportGlobalName}(dependencyModuleId)
        });

        pending[moduleId].promise = Promise.all(dependencyPromises).then(async function() {
            const url = window.${mashupResolveUrlGlobalName}(moduleId);
            await ${
              process.env.NODE_ENV === 'test' ? 'window.fakeImport' : 'import'
            }(url);
            
            return window.${mashupChunksGlobalName}[moduleId];
        });
    }
    return pending[moduleId].promise;
}).bind({});
`;
}

module.exports = function mashup(settings) {
  let publicApis = [];

  // options are taken from the object's property to
  // allow them to be overriden in plugins array
  const plugin = {
    name: PLUGIN_NAME,
    settings,
    generateBundle: async function(outputOptions, bundle) {
      // all regular chunks (isEntry: true) are loaded with our miniloader
      // but there are also commons chunks extracted automatically to avoid
      // duplication, and even our entry module may use these
      // so we need to inline them with our entry point chunk
      const {
        entryPoint,
        mashupChunksGlobalName,
        mashupDependenciesGlobalName,
        mashupDependencyMapGlobalName,
        mashupResolveUrlGlobalName,
        mashupIntegrationName,
        mashupImportGlobalName,
        mashupModuleName,
        mashupInitializeFunction,
        mashupGlobalName,
        basePath,
        target,
        publicApiEntry,
        dependencies
      } = plugin.settings;

      const contentEvalFunctionName = `${mashupGlobalName}_evaluate`;

      let entryPointChunk;
      let publicApiChunkName;
      let publicApiChunk;

      const dependencyMap = {};
      for (const [chunkName, chunk] of Object.entries(bundle)) {
        if (chunk.modules[entryPoint]) {
          entryPointChunk = chunk;
        } else if (publicApiEntry && chunk.modules[publicApiEntry]) {
          publicApiChunkName = chunkName;
          publicApiChunk = chunk;
        } else {
          const {imports} = chunk;
          dependencyMap[`./${chunkName}`] = [];
          for (const importName of imports) {
            if (bundle[importName]) {
              dependencyMap[`./${chunkName}`].push(importName);
            }
          }
        }
      }

      if (Object.keys(dependencyMap).length > 0) {
        console.log(
          ` - async chunk dependency tree: ${cyan(
            JSON.stringify(dependencyMap)
          )}`
        );
      }

      const concat = new Concat(true, entryPoint);
      concat.add(
        null,
        getMashupModuleLoader({
          mashupChunksGlobalName,
          mashupDependenciesGlobalName,
          mashupImportGlobalName,
          mashupDependencyMapGlobalName,
          dependencyMap,
          mashupResolveUrlGlobalName,
          mashupIntegrationName,
          basePath,
          mashupGlobalName,
          dependencies
        })
      );

      concat.add(
        null,
        `function ${contentEvalFunctionName}(deps) { 
          if (window.${mashupGlobalName}_evaluated) return;
          window.${mashupGlobalName}_evaluated = true;
          ${dependencies
            .map(
              (item, index) =>
                `window.${mashupDependenciesGlobalName}['${item.provideModule ||
                  item.mashupDependency ||
                  item}'] = deps[${index}];`
            )
            .join('\n')};
      `
      );

      // If some of the commons chunks are imported by entry point synchronously
      // (i.e. not using import()) we'll have to pre-load them before evaluating entry point
      // anyway - we can avoid extra requests by inlining commons chunks directly
      // and since public API chunk is also inlined with entry point, its direct imports
      // need to be inlined here as well. If entry point and public API synchronously import
      // the same chunks, we only need one copy of each, so Set here is used do deduplicate.
      const entryPointImports = new Set([
        ...entryPointChunk.imports,
        ...((publicApiChunk && publicApiChunk.imports) || [])
      ]);

      for (const importName of entryPointImports) {
        if (bundle[importName]) {
          const {code, map} = bundle[importName];
          concat.add(importName, code, JSON.stringify(map));
          delete bundle[importName];
        }
      }
      concat.add(
        entryPointChunk.fileName,
        entryPointChunk.code,
        JSON.stringify(entryPointChunk.map)
      );

      if (publicApiChunkName) {
        const {code, map} = bundle[publicApiChunkName];
        concat.add(publicApiChunkName, code, JSON.stringify(map));
        delete bundle[publicApiChunkName];
      }

      concat.add(
        null,
        `window.${mashupGlobalName} = window.${mashupChunksGlobalName}['./${entryPointChunk.fileName}'];}`
      );

      for (const chunkName of Object.keys(bundle)) {
        // if there are no dynamically-loaded chunks, basePath is not required
        // but if there are some chunks left that are marked as non-entry,
        // we require baseBath to know where to dynamically load them from
        if (!bundle[chunkName].isEntry && !basePath) {
          const filenames = Object.keys(bundle[chunkName].modules);
          throw new Error(
            `basePath not specified, but the following files are loaded with dynamic import(): ${filenames
              .map(filename => ` - ${cyan(filename)}`)
              .join(
                '\n'
              )}Code splitting requires basePath and integrationName to correctly resolve dynamically loaded chunk URLs.`
          );
        }
      }

      publicApis.forEach(api => {
        concat.add(
          null,
          `window.${mashupGlobalName}_getDependencies().addModule('${api}',
              function () { 
                ${contentEvalFunctionName}(arguments);
                return window.${mashupChunksGlobalName}['./${publicApiChunkName}']['${api}'](); 
              });`
        );
      });

      concat.add(
        null,
        `window.${mashupGlobalName}_getDependencies()
      .${
        target === 'mashup' ? 'addMashup(' : `addModule('${mashupModuleName}',`
      }
  function() {
${contentEvalFunctionName}(arguments);
${
  target === 'mashup'
    ? `window.${mashupInitializeFunction}();`
    : `return {initialize: window.${mashupInitializeFunction}};`
}
});`
      );

      entryPointChunk.code = concat.content.toString();
      if (outputOptions.sourcemap) {
        entryPointChunk.map.mappings = JSON.parse(concat.sourceMap).mappings;
      }

      if (target === 'mashup') {
        bundle['manifest.cfg'] = {code: 'Placeholders:restui_board'};
      }
    },
    renderChunk: async function(sourceCode, options) {
      const {
        entryPoint,
        basePath,
        dependencies,
        target,
        mashupModuleName,
        mashupGlobalName,
        mashupImportGlobalName,
        mashupInitializeFunction,
        mashupChunksGlobalName,
        mashupDependenciesGlobalName,
        publicApiEntry
      } = plugin.settings;

      // our global names are window.<something>_chunks['<chunk id>']
      // but iife plugin treats those as variable names and prepends var
      // we insert spaces instead of var here to preserve symbol position
      // so that we don't have to merge sourcemaps
      sourceCode = sourceCode.split('var window.').join('    window.');

      const code = new MagicString(sourceCode);
      const ast = this.parse(code.toString(), {
        // we also need to remove all comments because require('<something>') in comments
        // causes TP SystemJS loader to treat it as external module to be loaded
        onComment(isBlock, text, start, end) {
          code.overwrite(start, end, '');
        }
      });
      // if chunk contains entry point module, it is an entry point chunk
      if (options.modules[entryPoint]) {
        const [initStatement] = ast.body;
        // window.<something> = <right>, RHS is IIFE for entry point
        // we need to replace it with regular function, because it can only be called after
        // all mashup dependencies are assigned properly
        const assignmentExpression = initStatement.expression;
        if (assignmentExpression.type !== 'AssignmentExpression') {
          throw new Error(
            `Mashup entry point module ${entryPoint} doesn't export anything. It must have a named export function 'initialize'. Any toplevel code using dependencies before 'initialize' call is invalid.`
          );
        }

        // save entry function to global variable for further call from addModule function
        walk(ast, {
          enter(node, parent) {
            if (
              (node.type === 'FunctionDeclaration' ||
                node.type === 'VariableDeclarator') &&
              node.id.name === 'initialize' &&
              !node.parent
            ) {
              code.appendRight(
                node.end,
                `;window.${mashupInitializeFunction} = initialize;`
              );
            }
          }
        });
      }

      if (publicApiEntry && options.modules[publicApiEntry]) {
        walk(ast, {
          enter(node, parent) {
            if (
              node.type === 'ObjectExpression' &&
              parent &&
              parent.type === 'VariableDeclarator'
            ) {
              publicApis = node.properties.map(property => property.key.value);
            }
          }
        });
      }

      // regardless of module type, we need to replace dynamic imports
      // with our own import function, and naive string replace could
      // malform strings/comments/etc.
      walk(ast, {
        enter(node, parent) {
          if (node.type === 'ImportExpression') {
            // node.start and node.end correspond to 6 characters - 'import'
            // we replace these with a regular function call - <something>_import
            code.overwrite(node.start, node.start + 6, mashupImportGlobalName);
          }
        }
      });

      return {
        code: code.toString(),
        map: code.generateMap({hires: true})
      };
    }
  };

  return plugin;
};

module.exports.pluginName = PLUGIN_NAME;
